program s
  use input_mdl, only: muUV,muIR,muMS,dp
  use hs_ct, only: initct
  use fill_couplings, only: fill_couplings_mdl
  use lbfgsb_interface, only: chi2func_ptr,dchi2func_ptr,init_grid,set_bounds,run_loop
  use modelinput, only: linc,printp,rs_ref,rs_cmp
  use class_particles, only: print_parameters_mdl
  implicit none

  procedure(chi2func_ptr), pointer :: chi2f
  procedure(dchi2func_ptr), pointer :: dchi2f
  real(dp) :: ref_a0,ref_l30,a,b,chires,sa_in,l3_in
  real(dp), parameter :: sol_thr = 1d-13

  character(len=*), parameter :: fmt_s = '(2x,a)'
  character(len=*), parameter :: fmt = '(2x,a,x,f23.14)'
  character(len=*), parameter :: fmt_a = '(2x,a,x,a)'

  chi2f => chi2_hs
  dchi2f => dchi2_hs

  call read_input

  if (printp .gt. 0) call print_parameters_mdl(printp-1)

  if (linc .eq. 0) then
    call run_loop
    call write_result
  else
    call lin_conversion
    call write_linc_result
  end if

  contains  

  subroutine write_result
    use class_particles, only: sa,l3
    write(*,*)
    write(*,fmt_a) "Full conversion from " // trim(rs_ref) // " to " // trim(rs_cmp)

    call chi2_hs([real(l3,kind=dp),real(sa,kind=dp)],chires)
    if (chires .gt. sol_thr) then
      write(*,*) "Only approximate solution."
    end if
    write(*,fmt) "    sa:", real(sa)
    write(*,fmt) "    ca:", sqrt(1d0-real(sa)**2)
    write(*,fmt) "    l3:", real(l3)
    write(*,fmt) "chires:", chires
    write(*,*)
  end subroutine write_result

  subroutine write_linc_result
    use class_particles, only: sa,l3
    write(*,*)
    write(*,fmt_a) "Linear conversion from " // trim(rs_ref) // " to " // trim(rs_cmp)
    write(*,fmt) "    sa:", real(sa)
    write(*,fmt) "    ca:", sqrt(1d0-real(sa)**2)
    write(*,fmt) "    l3:", real(l3)
    write(*,fmt) "chires:", 0.
    write(*,*)
  end subroutine write_linc_result

  subroutine read_input
    use fill_couplings, only: fill_couplings_mdl
    use hs_ct, only: initct
    use collier_interface, only: init_collier
    use modelinput, only: read_modelinput,sa,l3,MHH,printp,mumsbsm
    use modelfile, only: set_parameter_mdl
    character(len=32) :: arg
    integer :: count
    real(dp) :: mhh_in,mha_in,mhc_in,msb_in,massreg
    complex(dp),parameter :: cone=1d0,cnul=0d0

    call read_modelinput

    muUV = 100d0
    muIR = 100d0
    muMS = 100d0

    mhh_in = MHH
    sa_in = sa
    l3_in = l3
    if (printp .eq. 0) then
      write(*,fmt) "mhh in:", mhh_in
      write(*,fmt) " sa in:", sa_in
      write(*,fmt) " l3 in:", l3_in
    end if

    call set_parameter_mdl('sa', sa_in*cone)
    call set_parameter_mdl('MHH', mhh_in*cone)
    call set_parameter_mdl("WHH",cnul)
    call set_parameter_mdl('l3', l3_in*cone)

    if (mumsbsm .eq. -1d0) then
      muMS = 125d0
    else
      muMS = mumsbsm
    end if
    muUV = muMS
    call set_parameter_mdl('muMS_BSM', muMS*cone)

    call init_collier
    call fill_couplings_mdl
    call initct(0d0)
    call set_ref_values
    call init_grid(2,10)
    call set_initial_bounds

  end subroutine read_input

  subroutine set_initial_bounds()
    call set_bounds([-abs(l3_in)*4, -1d0], &
                    [abs(l3_in)*4, 1d0],  &
                    [l3_in, sa_in],  &
                    chi2f,dchi2f)
    
  end subroutine set_initial_bounds

  subroutine lin_conversion
    use class_particles, only: sa,l3
    real(dp)             :: a,an,sar

    sar=sa
    a = asin(sar)
    an = a + get_da(rs_ref) - get_da(rs_cmp)
    l3 = l3 + get_dl3(rs_ref) - get_dl3(rs_cmp)
    sa = sin(an)
    chires = 0
  end subroutine lin_conversion

  function get_da(scheme)

    use hs_ct, only: da_FJTS, da_PRTS, da_OS, da_BFM
    character(len=*) :: scheme
    real(dp)         :: get_da

    select case (scheme)
      case ('FJ')
        get_da = da_FJTS
      case ('PR')
        get_da = da_PRTS
      case ('OS')
        get_da = da_OS
      case ('BFM')
        get_da = da_BFM
      case default
        write(*,*) "Requested RS: ",scheme
        write(*,*) "but not available."
        write(*,*) "Allowed schemes: FJ, PR, OS, BFM"
        stop 9
        
    end select 
    
  end function get_da

  function get_dl3(scheme)
    use hs_ct, only: dl3_l1_MS_OS, dl3_l1_MS_FJTS, &
                     dl3_l1_MS_PRTS, dl3_BFM
    character(len=*) :: scheme
    real(dp)         :: get_dl3

    select case (scheme)
      case ('FJ')
        get_dl3 = dl3_l1_MS_FJTS
      case ('PR')
        get_dl3 = dl3_l1_MS_PRTS
      case ('OS')
        get_dl3 = dl3_l1_MS_OS
      case ('BFM')
        get_dl3 = dl3_BFM
      case default
        write(*,*) "Requested RS: ",scheme
        write(*,*) "but not available."
        write(*,*) "Allowed schemes: FJTS, PRTS, OS, BFM"
        stop 9
        
    end select 
    
  end function get_dl3

  subroutine set_ref_values
    use class_particles, only: sa,l3
    real(dp) :: a,sar

    sar=sa
    a = asin(sar)
    ref_a0 = a + get_da(rs_ref)
    ref_l30 = l3 + get_dl3(rs_ref)
  end subroutine set_ref_values

  subroutine chi2_hs(x,res)
    use hs_ct, only: initct
    use class_particles, only: l3,sa,init_parameters_mdl
    use fill_couplings, only: fill_couplings_mdl
    real(dp), dimension(:), intent(in)  :: x
    real(dp),               intent(out) :: res
    real(dp) :: a,sar
    l3 = x(1)
    sa = x(2)
    sar=sa
    a = asin(sar)
    call init_parameters_mdl
    call fill_couplings_mdl()
    call initct(0d0)
    ! TODO:  <29-11-18, J.-N. Lang> !
    ! make l3 dimensionless 
    res = (ref_l30 - l3 - get_dl3(rs_cmp))**2 + (ref_a0 - a - get_da(rs_cmp))**2
  end subroutine chi2_hs

  subroutine dchi2_hs(dim,x,df)

    integer,                  intent(in)  :: dim
    real(dp), dimension(dim), intent(in)  :: x 
    real(dp), dimension(dim), intent(out) :: df
    real(dp) :: fval,yp,ym
    real(dp), parameter :: eps=1d-8
    real(dp), dimension(dim) :: x_tmp
    integer :: i

    call chi2_hs(x,fval)
    x_tmp = x
    do i = 1, dim
      x_tmp(i) = x(i)+eps
      call chi2_hs(x_tmp,yp)
      x_tmp(i) = x(i)-eps
      call chi2_hs(x_tmp,ym)
      df(i) = (yp-ym)/2d0/eps
      x_tmp(i) = x(i)
    end do
   
  end subroutine dchi2_hs

end program s
